Set Automatic Coercions Import.

Require Import ssreflect ssrfun ssrbool ssrnat seq prefix.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(******************************************************************************)
(*       Module for searching and inserting elements in a list                *)
(******************************************************************************)

Section XFind.

Variable A : Type.

Definition invariant s r i (e : A) := onth r i = Some e /\ prefix s r.

(* Tagging for controlling the search of instances *)
Structure xtagged := XTag {xuntag :> A}.

Definition extend_tag := XTag.
Definition recurse_tag := extend_tag.
Canonical Structure found_tag x := recurse_tag x.

(* Main structure
   s : input sequence
   r : output sequence. If elem_of is in the sequence, then it's equal to s, 
       otherwise it's equal to (elem_of :: s)
   i : output index of elem_of in r *)
Structure xfind (s r : seq A) (i : nat) := XFind {
  elem_of :> xtagged; 
  _ : invariant s r i elem_of}.

Implicit Arguments XFind [].


Lemma found_pf x t : invariant (x :: t) (x :: t) 0 x. 
Proof. by split; [|apply: prefix_refl]. Qed.

Canonical Structure found_struct x t := 
  XFind (x :: t) (x :: t) 0 (found_tag x) (found_pf x t).

Lemma recurse_pf (i : nat) (y : A) (s r : seq A) (f : xfind s r i) :
        invariant (y :: s) (y :: r) i.+1 f.
Proof. by case:f=>[q [H1 H2]]; split; [|apply/prefix_cons]. Qed.

Canonical Structure recurse_struct i y t r (f : xfind t r i) :=
  XFind (y :: t) (y :: r) i.+1 (recurse_tag f) (recurse_pf y f).

Lemma extend_pf x : invariant [::] [:: x] 0 x.
Proof. by []. Qed.

Canonical Structure extend_struct x := 
  XFind [::] [:: x] 0 (extend_tag x) (extend_pf x).

End XFind.

Lemma findme A (r s : seq A) i (f : xfind r s i) : onth s i = Some (xuntag (elem_of f)).
by case: f=>e [/= ->].
Qed.

Example test A (x1 x2 x3 : A) : onth [:: x1; x2; x3] 2 = Some x3.
apply: findme.
Defined.

Set Printing Implicit.
Print test.

Example unit_test : forall A (x1 x2 x3 x y : A),
   (forall s r i (f : xfind s r i), nth x1 r i = xuntag f -> xuntag f = x) ->
  x = x.
Proof.
move=>A x1 x2 x3 x y test_form.
apply: (test_form [::]). simpl.
apply: (test_form [:: x1; x]). simpl.
apply: (test_form [:: x1; x2; x; x3]). simpl.
apply: (test_form [:: x1; x2; x3]). simpl.
Abort.

Example stress  
  (x01 x02 x03 x04 x05 x06 x07 x08 x09 x10 : nat) 
  (x11 x12 x13 x14 x15 x16 x17 x18 x19 x20 : nat) 
  (x21 x22 x23 x24 x25 x26 x27 x28 x29 x30 : nat) 
  (x31 x32 x33 x34 x35 x36 x37 x38 x39 x40 : nat) 
  (x41 x42 x43 x44 x45 x46 x47 x48 x49 x50 : nat) 
  (H : forall r i
  (f : xfind ([:: x01; x02; x03; x04; x05; x06; x07; x08; x09; x10] ++
    [:: x11; x12; x13; x14; x15; x16; x17; x18; x19; x20] ++
    [:: x21; x22; x23; x24; x25; x26; x27; x28; x29; x30] ++
    [:: x31; x32; x33; x34; x35; x36; x37; x38; x39; x40] ++
    [:: x01; x42; x43; x44; x45; x46; x47; x48; x49]) r i),
  xuntag f = x50 -> i = 49)
  : 49 = 49.
Time apply: (H _ _ _ (erefl _)).
Time Qed.